-module(rumHWatcherSrv).

-behaviour(gen_srv).

-include("eRum.hrl").

-export([
   start/3
   , start_link/3
]).

-export([
   init/1
   , handleCall/3
   , handleCast/2
   , handleInfo/2
   , terminate/2,
   code_change/3
]).

-record(state, {
   module :: atom(),
   config :: any(),
   sink :: pid() | atom()
}).

start(Sink, Module, Config) ->
   gen_srv:start(?MODULE, {Sink, Module, Config}, []).

start_link(Sink, Module, Config) ->
   gen_srv:start_link(?MODULE, {Sink, Module, Config}, []).

init({Sink, Module, Config}) ->
   process_flag(trap_exit, true),
   installHandler(Module, Config, Sink),
   {ok, #state{sink = Sink, module = Module, config = Config}}.

handleCall(_Msg, _State, _From) ->
   ?ERR("~p call receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   {reply, ok}.

handleCast(_Msg, _State) ->
   ?ERR("~p cast receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   kpS.

handleInfo({gen_event_EXIT, Module, normal}, #state{module = Module} = State) ->
   {stop, normal, State};
handleInfo({gen_event_EXIT, Module, shutdown}, #state{module = Module} = State) ->
   {stop, normal, State};
handleInfo({gen_event_EXIT, Module, {'EXIT', {kill_me, [_KillerHWM, KillerReinstallAfter]}}}, #state{module = Module, sink = Sink, config = Config} = State) ->
   %% Brutally kill the manager but stay alive to restore settings.
   %%
   %% SinkPid here means the gen_event process. Handlers *all* live inside the
   %% same gen_event process space, so when the Pid is killed, *all* of the
   %% pending log messages in its mailbox will die too.
   SinkPid = whereis(Sink),
   unlink(SinkPid),
   {message_queue_len, Len} = process_info(SinkPid, message_queue_len),
   error_logger:error_msg("Killing sink ~p, current message_queue_len:~p~n", [Sink, Len]),
   exit(SinkPid, kill),
   _ = timer:apply_after(KillerReinstallAfter, lager_app, start_handler, [Sink, Module, Config]),
   {stop, normal, State};
handleInfo({gen_event_EXIT, Module, Reason}, #state{module = Module, config = Config, sink = Sink} = State) ->
   case eRum:log(error, self(), "Lager event handler ~p exited with reason ~s", [Module, rumErrLoggerH:format_reason(Reason)]) of
      ok ->
         installHandler(Module, Config, Sink);
      {error, _} ->
         %% lager is not working, so installing a handler won't work
         ok
   end,
   {noreply, State};
handleInfo(mReinstallHandler, #state{module = Module, config = Config, sink = Sink}) ->
   installHandler(Module, Config, Sink),
   kpS;
handleInfo({reboot, Sink}, _State) ->
   eRum_app:startSink(Sink),
   kpS;
handleInfo(stop, State) ->
   {stop, normal, State};
handleInfo({'EXIT', _Pid, killed}, #state{module = Module, config = Config, sink = Sink} = State) ->
   Tmr = rumUtil:get_env(killerReTime, 5000),
   _ = timer:apply_after(Tmr, eRum_app, startHandler, [Sink, Module, Config]),
   {stop, normal, State};
handleInfo(_Msg, _State) ->
   ?ERR("~p info receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   kpS.

terminate(_Reason, _State) ->
   ok.

code_change(_OldVsn, State, _Extra) ->
   {ok, State}.

installHandler(Module, Config, Sink) ->
   Ret =
      case Module of
         rumBackendThrottle ->
            %% lager_backend_throttle需要知道它连接到哪个接收器，因此这个公认的丑陋的解决方案。处理程序对发送到' init'的配置结构是敏感的，
            %% 遗憾的是，在不破坏第三方处理程序的情况下，向后端添加一个要忽略的配置项并不是一件小事。
            gen_emm:add_sup_epm(Sink, Module, [{sink, Sink} | Config]);
         _ ->
            gen_emm:add_sup_epm(Sink, Module, Config)
      end,

   case Ret of
      ok ->
         ?INT_LOG(debug, "Lager installed handler ~p into ~p", [Module, Sink]),
         eRum:updateLogevelCfg(Sink),
         ok;
      {error, {fatal, Reason}} ->
         ?INT_LOG(error, "Lager fatally failed to install handler ~p into ~p, NOT retrying: ~p", [Module, Sink, Reason]),
         %% tell ourselves to stop
         self() ! stop,
         ok;
      Error ->
         %% try to reinstall it later
         ?INT_LOG(error, "Lager failed to install handler ~p into ~p, retrying later : ~p", [Module, Sink, Error]),
         erlang:send_after(5000, self(), mReinstallHandler),
         ok
   end.

-ifdef(TEST).

from_now(Seconds) ->
   {Mega, Secs, Micro} = os:timestamp(),
   {Mega, Secs + Seconds, Micro}.

reinstall_on_initial_failure_test_() ->
   {timeout, 60000,
      [
         fun() ->
            error_logger:tty(false),
            application:load(lager),
            application:set_env(lager, handlers, [{lager_test_backend, info}, {lager_crash_backend, [from_now(2), undefined]}]),
            application:set_env(lager, errLoggerRedirect, false),
            application:unset_env(lager, crash_log),
            eRum:start(),
            try
               {_Level, _Time, Message, _Metadata} = lager_test_backend:pop(),
               ?assertMatch("Lager failed to install handler lager_crash_backend into lager_event, retrying later :" ++ _, lists:flatten(Message)),
               timer:sleep(6000),
               lager_test_backend:flush(),
               ?assertEqual(0, lager_test_backend:count()),
               ?assert(lists:member(lager_crash_backend, gen_event:which_handlers(rumEvent)))
            after
               application:stop(lager),
               application:stop(goldrush),
               error_logger:tty(true)
            end
         end
      ]
   }.

reinstall_on_runtime_failure_test_() ->
   {timeout, 60000,
      [
         fun() ->
            error_logger:tty(false),
            application:load(lager),
            application:set_env(lager, handlers, [{lager_test_backend, info}, {lager_crash_backend, [undefined, from_now(5)]}]),
            application:set_env(lager, errLoggerRedirect, false),
            application:unset_env(lager, crash_log),
            eRum:start(),
            try
               ?assert(lists:member(lager_crash_backend, gen_event:which_handlers(rumEvent))),
               timer:sleep(6000),

               pop_until("Lager event handler lager_crash_backend exited with reason crash", fun lists:flatten/1),
               pop_until("Lager failed to install handler lager_crash_backend into lager_event, retrying later",
                  fun(Msg) -> string:substr(lists:flatten(Msg), 1, 84) end),
               ?assertEqual(false, lists:member(lager_crash_backend, gen_event:which_handlers(rumEvent)))
            after
               application:stop(lager),
               application:stop(goldrush),
               error_logger:tty(true)
            end
         end
      ]
   }.

reinstall_handlers_after_killer_hwm_test_() ->
   {timeout, 60000,
      [
         fun() ->
            error_logger:tty(false),
            application:load(lager),
            application:set_env(lager, handlers, [{lager_manager_killer, [1000, 5000]}]),
            application:set_env(lager, errLoggerRedirect, false),
            application:set_env(lager, killerReTime, 5000),
            application:unset_env(lager, crash_log),
            eRum:start(),
            eRum:trace_file("foo", [{foo, "bar"}], error),
            L = length(gen_event:which_handlers(rumEvent)),
            try
               rumMgrKiller:kill_me(),
               timer:sleep(6000),
               ?assertEqual(L, length(gen_event:which_handlers(rumEvent))),
               file:delete("foo")
            after
               application:stop(lager),
               application:stop(goldrush),
               error_logger:tty(true)
            end
         end
      ]
   }.

pop_until(String, Fun) ->
   try_backend_pop(lager_test_backend:pop(), String, Fun).

try_backend_pop(undefined, String, _Fun) ->
   throw("Not found: " ++ String);
try_backend_pop({_Severity, _Date, Msg, _Metadata}, String, Fun) ->
   case Fun(Msg) of
      String ->
         ok;
      _ ->
         try_backend_pop(lager_test_backend:pop(), String, Fun)
   end.

-endif.
